
type OutputMapping{
  shopObjectType: String
  shopAttributeName: String
  cdfAttributeName: String
  unit: String
  isStep: Boolean
}

type OutputContainer{
  name: String
  watercourse: String
  shopType: String
  mappings: [OutputMapping]
}

type CommandConfig {
  commands: [String]
}

type Scenario {
  name: String
  template: ScenarioTemplate
  mapping: ScenarioMapping
  commands: CommandConfig
}


type ScenarioTemplate {
  watercourse: String
  shopVersion: String
  templateVersion: String
  model: File
  shopFiles: [File]
  baseMapping: ScenarioMapping
  outputDefinitions: OutputContainer
}
