type DayAheadBid{
    name: String
    date: [DateTransformation]
    market: NordPoolMarket
    isDefaultConfigForPriceArea: Boolean
    mainScenario: String
    priceArea: String
    watercourse: String
    priceScenarios: [ScenarioMapping]
    noShop: Boolean
    bidProcessConfigurationName: String
    bidMatrixGeneratorConfigExternalId: String
}

type BidMatrixGenerator {
    shopPlant: String
    methods: String
    functionExternalId: String
}

type DayAheadProcess{
    name: String
    bid: DayAheadBid
    shop: ShopTransformation
    bidMatrixGeneratorConfig: [BidMatrixGenerator]
    incremental_mappings: [ScenarioMapping]
}
